{{
  ----------------------------------------------
  Propeller Based Webserver
  ----------------------------------------------
}}

CON

  _clkmode = xtal1+pll16x
  _xinfreq = 5_000_000
  '''_xinfreq = 6_250_000

  rxlen = 128
  txlen = sdfat#SECTORSIZE

  TCPRXBUFLEN = 128
  TCPTXBUFLEN = 4096
  
OBJ

  web     : "api_telnet_serial"
  'text    : "VGA_Text_2"
  str     : "util_strings" 
  sdfat   : "fsrw24_rr003"

VAR

  'long keyStack[64]
  'byte keyBuffer[256]
  byte bSend
  long webStack[64]
  long size
  byte position 
  byte sendwebBuffer[txlen]
  byte receivewebBuffer[rxlen]
  'byte settings[17]
  byte temp
  byte fname[13]
  byte fending[4]
  long type
  byte reqstr[32]               ' request string

  byte tcpRxBuf[TCPRXBUFLEN]
  byte tcpTxBuf[TCPTXBUFLEN] 
  
  
DAT
  mac_addr      byte    $10, $00, $00, $00, $00, $01
  ip_addr       byte    192, 168, 178, 4                   ' device's ip address (CHANGE THIS)
  ip_subnet     byte    255, 255, 255, 0                   ' network subnet
  ip_gateway    byte    10, 10, 10, 254                    ' network gateway (router)
  ip_dns        byte    10, 10, 10, 254                    ' network dns
  dhcp_or_static byte   0

PUB start

  web.start(4,3,2,1,-1,-1,@mac_addr,@ip_addr)
  '''web.start(11, 10, 9, 8, -1, -1, @mac_addr, @ip_addr)
  if sdfat.mount_explicit(7,6,5,0) == 0
  '''ifnot \sdfat.mount(12)
     cognew(webServer, @webStack)

PRI webServer
  ' web server (runs in a seperate cog)
  \web.listen(80, @tcpRxBuf, TCPRXBUFLEN, @tcpTxBuf, TCPTXBUFLEN)
  repeat
    if \web.isConnected
      if \webServerThread == 0
        \web.txflush
      \web.close
    \web.relisten 

DAT

http200       byte      "HTTP/1.0 200 OK",13,10,0
http404       byte      "HTTP/1.0 404 Not Found",13,10,13,10,0
http501       byte      "HTTP/1.0 501 Not Implemented",13,10,13,10,0
http503       byte      "HTTP/1.0 503 Service Unavailable",13,10,13,10,0 
filelist      byte      "<html>Liste der Dateien auf dem Webserver:<br>------------------------------------<br><br>",0
crlf          byte      13,10,0

PRI webServerThread | ptr, row, c, h, j, n, m, i
  ' handles a single client connection
  m:=0
  if _webReadLine == 0                         ' read the first header, quit if it is empty
    return 0

  bytemove(@reqstr, @receivewebBuffer, 32)     ' copy the header to a temporary request string for later processing

  ' obtain get arguments
  {if (i := str.indexOf(@reqstr, string(".cgi?"))) <> -1 ' was the request for a *.cgi script with arguments?
    args := @reqstr[i + 5]                              ' extract the argument
    if (j := str.indexOf(args, string("="))) <> -1      ' find the end of the argument
      byte[args][j] := 0                                ' string termination
  }
  ' read the rest of the headers

  repeat until _webReadLine == 0              ' read the rest of the headers, throwing them away

  if str.indexOf(@reqstr, string("GET")) <> -1     
  elseif str.indexOf(@reqstr, string("POST")) <> -1  
  elseif str.indexOf(@reqstr, string("HEAD")) <> -1
    m:=1
  else
    web.str(@http501)
    return -1                   ' unrecognized method / not implemented; send 501 error message

  j:=0
  h:=0
  i:=0
  temp:=0

  position:=str.indexOf(@reqstr, string(" /"))+2
  repeat until j==12
    fname[j]:=reqstr[position]              ' reading the filename (it has to be in the 8.3 format or shorter)
    if fname[j]==32
     temp:=j
     fname[j+1]:=0                          ' let the filename be a zero terminated string (not necessary for fsrw, but avoiding problems when using strcomp)
     quit
    position++
    j++ 

  if fname[0]==32                           ' if there is no filename in the request, send a list of files which are stored on the sd card
    sdfat.opendir
    web.str(@http200)
    web.str(@crlf)
    web.str(@filelist)
    repeat until sdfat.nextfile(@fname)==-1
      temp++                                ' !!! Exception: I'm using the same variable for two different purposes (file length [see above] <-> number of files [here]) !!!
      web.str(@fname)
      web.str(string("<br>"))
    web.str(string("<br>"))
    web.str(string("Dateien insgesamt: "))
    web.dec(temp)
    web.str(string("</html>"))
    return 0

  
  if temp==0                                ' get the file extension...
    fending[0]:=fname[9]                    ' if the filename matches the 8.3 format, it's the last three entries of fname
    fending[1]:=fname[10]
    fending[2]:=fname[11]
  else 
    fending[0]:=fname[temp-3]               ' otherwise temp marks the space after the filename
    fending[1]:=fname[temp-2]
    fending[2]:=fname[temp-1]
  fending[3]:=0                             ' zero terminated string in order to use strcomp
  
  if strcomp(@fending, string("htm"))       ' ... and decide what MIME type it is
    type := string("text/html")             ' of course there are a lot more, but these
  elseif strcomp(@fending, string("avi"))   ' I think are the most common ones
    type := string("video/x-msvideo")       ' or at least the ones I needed ;-)
  elseif strcomp(@fending, string("mpg"))
    type := string("video/mpeg")
  elseif strcomp(@fending, string("jpg"))
    type := string("image/jpeg")
  elseif strcomp(@fending, string("gif"))
    type := string("image/gif")
  elseif strcomp(@fending, string("pdf"))
    type := string("application/pdf")
  elseif strcomp(@fending, string("zip"))
    type := string("application/zip")
  elseif strcomp(@fending, string("wav"))
    type := string("audio/x-wav")
  elseif strcomp(@fending, string("mp3"))
    type := string("audio/mpeg")
  elseif strcomp(@fending, string("m4a"))
    type := string("audio/x-m4a")
  else
    type := string(" ")

  'text.out($00)
  'web.str(@fname)

  if sdfat.popen(@fname,"r") <> 0          ' Try to open the requested file
    web.str(@http404)                      ' if it is not found on the sd card send back the 404 error message
  else                                       
    web.str(@http200)                      ' file has been found and will be transmitted,
    web.str(string("Server: PropServer"))  ' but first we are going to send the header
    web.str(@crlf)                         ' which includes some information about the server and file                                
    'web.str(string("Content-Length: "))
    'web.dec(size)
    'web.str(@crlf)
    web.str(string("Content-Type: "))
    web.str(type)
    web.str(@crlf)
    web.str(string("Connection: close"))
    web.str(@crlf)
    web.str(@crlf)                         ' This is the end of the Header 
    if m==1                                ' a HEAD request would stop here
      return 0
    else
      repeat                                                                
        h:=sdfat.pread(@sendwebBuffer,txlen)  ' now we are reading the first BufferSize (4096 default) bytes                         
        if h=<0                                   ' if there are no bytes to read then exit the loop                          
          quit                                                             
        else
          web.txdata(@sendwebBuffer,h)            ' send the h bytes read
          
  return 0

PRI _webReadLine | i, ch
  repeat i from 0 to 126
    ch := web.rxtime(500)
    if ch == 13
      ch := web.rxtime(500)
    if ch == -1 or ch == 10
      quit
    receivewebBuffer[i] := ch

  receivewebBuffer[i] := 0

  return i